
/* utility functions for the MVA algorithm		*/

#include <stdlib.h>
#include	<mex.h>
#include <math.h>
#include "mexMva.h"

/* global variables initialized from the MVAInit()	*/

/************ local variables used from the algorithm *************/
/* index vector for pointing the computing nodes */
static int *piIndex = NULL;

/* queue length intermediate */
static double	*pdNI = NULL;
/* where pdNI[k + iK*i] is the queue length for the service center k */
/* (0<=k<=iK), correspondig to the population piIndex, i can be   */
/* obtainded using the funcion ulMVALinearIndexes		  */

/******************************************************************/


/***************** IO of the algorithm ****************/
/* the input values of the mva algorithm */
static MVAIN	mvain = {
		0, 0, NULL, NULL, NULL};

/* the output values of the mva algorithm */
static MVAOUT	mvaout = {
		NULL, NULL, NULL, NULL, NULL};
/******************************************************/



/********************************************************/
/*		     LOCAL FUNCTIONS                    */
/********************************************************/
/*int iMVANIDim();
BOOL bMVALoop();
BOOL bMVAInitIndex();
unsigned long  ulMVALinearIndexes(int *piIndexToLin);
BOOL    bMVAAllocNI();*/


/* increments the vector piIndex (that is the vector	*/
/* containing the index to the node that the proc have	*/
/* to calculate) so that the we can explore all nodes	*/
/* return FALSE when the loop has to finish		*/
static BOOL bMVALoop()
{

	int i;

	for(i = 0 ; i < mvain.iR; i++)
		if (piIndex[i] == mvain.piN[i])
			piIndex[i] = 0;
		else 
		{
			piIndex[i]++;
			return TRUE;
		}

	return FALSE;

}

/* initialize the index vector according to rank of 	*/
/* of the current processor				*/
static BOOL bMVAInitIndex()
{
	int i;

	piIndex = (int*)malloc(sizeof(int)*mvain.iR);
	if(piIndex == NULL)
	{
		mexErrMsgTxt("The memory is exhausted !");
		return FALSE;
	}
	
	for(i = 0; i < mvain.iR; i++)
	 piIndex[i] = 0;

	return TRUE;
}

/* It traslates the address of node to compute in a      */
/* linear address in order to avoid the using of	*/
/* a multidimensional vector				*/
/* the procedure doesn't use the last index because     */
/* we reuse the same array when the last index is      */
/* incremented, this makes spare of memory space		*/
/* piIndexiLin is the vector of indexes                 */
static unsigned long  ulMVALinearIndexes(int *piIndexToLin)
{
	unsigned long  ulRet = 0; /* linearization value */
	int		       i;     /* index variable */

	for (i = mvain.iR - 2; i >= 0 ; i--)
		ulRet = ulRet * (mvain.piN[i] + 1) + piIndexToLin[i];

	return ulRet;
}

/* it computes the number of items to run for a process */
/* we suppose that the population of the class 0 is     */
/* divided among the processors 			*/
static int iMVANIDim()
{

	int iReturnValue = 1;
	int i;

	for(i = 0; i < mvain.iR - 1; i++)
		iReturnValue = iReturnValue * (mvain.piN[i] + 1);

	return iReturnValue;
}


/* it allocs the space needed for storing the queue length */
static BOOL	bMVAAllocNI()
{
	pdNI = (double*)malloc(sizeof(double)*iMVANIDim()*mvain.iK);
	if(pdNI == NULL)
	{
		mexErrMsgTxt("The memory is exhausted !");
		return FALSE;
	}

	return TRUE;
}


/* it allocs the structure mvaout			*/
static BOOL bMVAAllocOUT()
{

	mvaout.pdX = (double*)malloc(sizeof(double)*mvain.iR);
	if(mvaout.pdX == NULL)
	{
		mexErrMsgTxt("The memory is exhausted !");
		return FALSE;
	}
	mvaout.pdU = (double*)malloc(sizeof(double)*mvain.iR*mvain.iK);
	if(mvaout.pdU == NULL)
	{
		mexErrMsgTxt("The memory is exhausted !");
		return FALSE;
	}
	mvaout.pdQ = (double*)malloc(sizeof(double)*mvain.iR*mvain.iK);
	if(mvaout.pdQ == NULL)
	{
		mexErrMsgTxt("The memory is exhausted !");
		return FALSE;
	}
	mvaout.pdR = (double*)malloc(sizeof(double)*mvain.iR*mvain.iK);
	if(mvaout.pdR == NULL)
	{
		mexErrMsgTxt("The memory is exhausted !");
		return FALSE;
	}
	mvaout.pdRC = (double*)malloc(sizeof(double)*mvain.iR);
	if(mvaout.pdRC == NULL)
	{
		mexErrMsgTxt("The memory is exhausted !");
		return FALSE;
	}

	return TRUE;
}


/********************************************************/
/*           INTERFACE FUNCTIONS			*/
/********************************************************/

/* init the environment for the PMVA algorithm		*/

BOOL bMVAInit()
{
	if(!bMVAInitIndex() || !bMVAAllocNI() || !bMVAAllocOUT())
		return FALSE;
	else
		return TRUE;
}


/* release the environment for the PMVA algorithm		*/
void MVARelease()
{

	MVAFreeIN(mvain);
	free(piIndex);
	free(pdNI);
	free(mvaout.pdX);
	free(mvaout.pdU);
	free(mvaout.pdQ);
	free(mvaout.pdR);
	free(mvaout.pdRC);
}


/* THE MVA Algorithm !				      	*/
/* piN: classes population vector                       */
/* iR: number of classes                                */
/* piIndex is the vector of indexes                     */
void MVAAlgo()
{

	int c, k;

	unsigned long ulCurrInd, ulGetInd;

	/* the main loop of the algorithm */

	do
	{
		ulCurrInd = ulMVALinearIndexes(piIndex);

		/* we compute here the queue lenght, of the current piIndex node */

		/***************** MVA ALGORITHM ****************/


		for(c = 0; c < mvain.iR; c++)
			if(piIndex[c] != 0)
			{
				piIndex[c]--;
				ulGetInd = ulMVALinearIndexes(piIndex);
				piIndex[c]++;

				mvaout.pdRC[c] = 0.0;

				for(k = 0; k < mvain.iK; k++)
				{
					mvaout.pdR[c + k*mvain.iR] = mvain.pdD[c + k*mvain.iR]*(1. + mvain.pbQueue[k]*pdNI[k + ulGetInd*mvain.iK]);

					mvaout.pdRC[c] += mvaout.pdR[c + k*mvain.iR];
				}

				mvaout.pdX[c] = ((double)piIndex[c])/mvaout.pdRC[c];
			}
		else
			{
				mvaout.pdRC[c] = 0.0;

				mvaout.pdX[c] = 0.0;

				for(k = 0; k < mvain.iK; k++)
					mvaout.pdR[c + k*mvain.iR] = 0.0;
			}


		for(k = 0; k < mvain.iK; k++)
		{
			pdNI[k + ulCurrInd*mvain.iK] = 0.0;

			for(c = 0; c < mvain.iR; c++)
				pdNI[k + ulCurrInd*mvain.iK] += mvaout.pdX[c]*mvaout.pdR[c + k*mvain.iR];
		}

		/******************** MVA END ********************/

	}while(bMVALoop());

	for(k = 0; k < mvain.iK; k++)
		for(c = 0; c < mvain.iR; c++)
		{
			mvaout.pdQ[c + k*mvain.iR] =  mvaout.pdX[c]*mvaout.pdR[c + k*mvain.iR];
			mvaout.pdU[c + k*mvain.iR] =  mvaout.pdX[c]*mvain.pdD[c + k*mvain.iR];

		}
}

void MVAFreeIN()
{
	free(mvain.piN);

	free(mvain.pbQueue);

	free(mvain.pdD);
}


#define VECTOR_N0 prhs[0]
#define MATRIX_LQ prhs[1]
#define MATRIX_LD prhs[2]
#define MATRIX_OUT plhs[0]
#define TH_OUT plhs[1]


void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[] )
{
	unsigned int col, row;

	if (nlhs!=2) 
		mexErrMsgTxt("mexmva requires 2 output arguments.");

	if (nrhs!=3) 
		mexErrMsgTxt("mexmva requires 3 input arguments, \n N0, Lq and Ld.");

	if (mxGetM(VECTOR_N0)!=1 || mxIsChar(VECTOR_N0) || mxIsSparse(VECTOR_N0) || mxIsComplex(VECTOR_N0)) 
		mexErrMsgTxt("N0 argument must be \n a 1xN, full, numeric, real-valued matrix.");

	if ((mxGetN(MATRIX_LQ) != mxGetN(VECTOR_N0) || (mxGetN(MATRIX_LQ) != mxGetN(MATRIX_LD)))) 
		mexErrMsgTxt("The matrix N0, Lq and Ld must have the same number of columns, real-valued vector.");

	if (mxIsChar(MATRIX_LQ) || mxIsSparse(MATRIX_LQ) || mxIsComplex(MATRIX_LQ)) 
		mexErrMsgTxt("Lq argument must be \n a NxM, full, numeric, real-valued matrix.");

	if (mxIsChar(MATRIX_LD) || mxIsSparse(MATRIX_LD) || mxIsComplex(MATRIX_LD)) 
		mexErrMsgTxt("Ld argument must be \n a NxM, full, numeric, real-valued matrix.");

	mvain.iR = mxGetN(MATRIX_LD);
	mvain.iK = mxGetM(MATRIX_LD)+mxGetM(MATRIX_LQ);

	mvain.piN = (int*)malloc(sizeof(int)*mvain.iR);
	if(mvain.piN == NULL)
	{
		mexErrMsgTxt("The memory is exhausted !");
		return;
	}
	for(col=0;col<mvain.iR;col++)
		mvain.piN[col] = (mxGetPr(VECTOR_N0))[col];


	mvain.pbQueue = (BOOL*)malloc(sizeof(BOOL)*mvain.iK);
	if(mvain.pbQueue == NULL)
	{
		mexErrMsgTxt("The memory is exhausted !");
		return;
	}
	for(row=0; row<mxGetM(MATRIX_LQ); row++)
		mvain.pbQueue[row] = TRUE;

	for(row=mxGetM(MATRIX_LQ); row<mvain.iK; row++)
		mvain.pbQueue[row] = FALSE;


	mvain.pdD = (double*)malloc(sizeof(double)*mvain.iR*mvain.iK);
	if(mvain.pdD == NULL)
	{
		mexErrMsgTxt("The memory is exhausted !");
		return;
	}
	for (col=0; col < mvain.iR; col++) 
		for (row=0; row < mxGetM(MATRIX_LQ); row++) 
			mvain.pdD[col + mvain.iR*row] = (mxGetPr(MATRIX_LQ))[row+col*mxGetM(MATRIX_LQ)];
	
	for (col=0; col < mvain.iR; col++)
		for (row=mxGetM(MATRIX_LQ); row < mvain.iK; row++) 
			mvain.pdD[col + mvain.iR*row] = (mxGetPr(MATRIX_LD))[row-mxGetM(MATRIX_LQ)+col*mxGetM(MATRIX_LD)];
	

	if(!bMVAInit())
		return;

	MVAAlgo();

	MATRIX_OUT = mxCreateDoubleMatrix(mvain.iK, mvain.iR, mxREAL);
	
	TH_OUT = mxCreateDoubleMatrix(1, mvain.iR, mxREAL); 

	for (col=0; col < mvain.iR; col++) 
		for (row=0; row < mvain.iK; row++) 
			(mxGetPr(MATRIX_OUT))[row+col*mxGetM(MATRIX_OUT)] = mvaout.pdR[col + mvain.iR*row];

	for (col=0; col < mvain.iR; col++)
		(mxGetPr(TH_OUT))[col*mxGetM(TH_OUT)] = mvaout.pdX[col];



	MVARelease();

}
